@@ -17,7 +17,6 @@
 
 #include "server_config_util.h"
 
-#include <crypt.h>
 #include <errno.h>
 #include <stdarg.h>
 #include <stdio.h>
@@ -502,12 +501,8 @@ _nc_server_config_add_ssh_user_password(const struct ly_ctx *ctx, const char *tr
 
     NC_CHECK_ARG_RET(NULL, ctx, tree_path, password, config, 1);
 
-    cdata = (struct crypt_data *) calloc(sizeof(struct crypt_data), 1);
-    if (cdata == NULL) {
-        ERR(NULL, "Allocation of crypt_data struct failed.");
-        ret = 1;
-        goto cleanup;
-    }
+    cdata = calloc(1, sizeof *cdata);
+    NC_CHECK_ERRMEM_GOTO(!cdata, ret = 1, cleanup);
 
     hashed_pw = crypt_r(password, salt, cdata);
     if (!hashed_pw) {
