@@ -383,6 +383,9 @@ nc_server_tls_verify_cb(int preverify_ok, X509_STORE_CTX *x509_ctx)
             /* copy the client cert */
             data->session->opts.server.client_cert = X509_dup(cert);
             NC_CHECK_ERRMEM_RET(!data->session->opts.server.client_cert, 0);
+
+            /* verification was successful, override the in-built verification result */
+            X509_STORE_CTX_set_error(x509_ctx, X509_V_OK);
         }
         return 1;
     } else {
@@ -404,10 +407,38 @@ nc_server_tls_set_verify_wrap(void *tls_cfg, struct nc_tls_verify_cb_data *cb_da
     SSL_CTX_set_verify(tls_cfg, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, nc_server_tls_verify_cb);
 }
 
-void
+int
 nc_client_tls_set_verify_wrap(void *tls_cfg)
 {
+    int ret = 0;
+    X509_VERIFY_PARAM *vpm = NULL;
+
+    /* set the verify flag */
     SSL_CTX_set_verify(tls_cfg, SSL_VERIFY_PEER, NULL);
+
+    vpm = X509_VERIFY_PARAM_new();
+    NC_CHECK_ERRMEM_RET(!vpm, 1);
+
+    /* set the partial chain flag to allow verification of a certificate chain
+     * to succeed even if the chain is not complete.
+     * See https://github.com/openssl/openssl/issues/7871
+     * This is not set for the server, because all the CA certs in the chain
+     * may be needed for CTN, so such partial chain cases are handled manually. */
+    if (!X509_VERIFY_PARAM_set_flags(vpm, X509_V_FLAG_PARTIAL_CHAIN)) {
+        ERR(NULL, "Setting X509_V_FLAG_PARTIAL_CHAIN flag failed (%s).", ERR_reason_error_string(ERR_get_error()));
+        ret = 1;
+        goto cleanup;
+    }
+
+    if (!SSL_CTX_set1_param(tls_cfg, vpm)) {
+        ERR(NULL, "Failed to set verify param (%s).", ERR_reason_error_string(ERR_get_error()));
+        ret = 1;
+        goto cleanup;
+    }
+
+cleanup:
+    X509_VERIFY_PARAM_free(vpm);
+    return ret;
 }
 
 char *
@@ -774,7 +805,7 @@ nc_tls_move_crls_to_store(const X509_STORE *src, X509_STORE *dst)
 }
 
 int
-nc_tls_setup_config_from_ctx_wrap(struct nc_tls_ctx *tls_ctx, int side, void *tls_cfg)
+nc_tls_setup_config_from_ctx_wrap(struct nc_tls_ctx *tls_ctx, void *tls_cfg)
 {
     if (SSL_CTX_use_certificate(tls_cfg, tls_ctx->cert) != 1) {
         ERR(NULL, "Setting up TLS certificate failed (%s).", ERR_reason_error_string(ERR_get_error()));
@@ -786,11 +817,6 @@ nc_tls_setup_config_from_ctx_wrap(struct nc_tls_ctx *tls_ctx, int side, void *tl
         return 1;
     }
 
-    /* disable server-side automatic chain building */
-    if (side == NC_SERVER) {
-        SSL_CTX_set_mode(tls_cfg, SSL_MODE_NO_AUTO_CHAIN);
-    }
-
     if (tls_ctx->crl_store) {
         /* move CRLs from crl_store to cert_store, because SSL_CTX can only have one store */
         if (nc_tls_move_crls_to_store(tls_ctx->crl_store, tls_ctx->cert_store)) {
