@@ -5,7 +5,7 @@
  * @brief libnetconf2 session manipulation
  *
  * @copyright
- * Copyright (c) 2017 - 2023 CESNET, z.s.p.o.
+ * Copyright (c) 2017 - 2025 CESNET, z.s.p.o.
  *
  * This source code is licensed under BSD 3-Clause License (the "License").
  * You may not use this file except in compliance with the License.
@@ -39,6 +39,82 @@
 
 extern struct nc_server_opts server_opts;
 
+/**
+ * Size added to the reader buffer on realloc.
+ * Used only for NETCONF 1.0 messages (always the hello message).
+ */
+#define NC_READ_BUF_SIZE_STEP 512
+
+/**
+ * Maximum size of a framing message chunk (spec max is 2^32 - 1).
+ */
+#define NC_WRITE_CHUNK_SIZE_MAX 1024 * 10
+
+/**
+ * Sleep time in usec to wait between nc_recv_notif() calls.
+ */
+#define NC_CLIENT_NOTIF_THREAD_SLEEP 10000
+
+/**
+ * Timeout in msec for transport-related data to arrive (ssh_handle_key_exchange(), SSL_accept(), SSL_connect()).
+ * It can be quite a lot on slow machines (waiting for TLS cert-to-name resolution, ...).
+ */
+#define NC_TRANSPORT_TIMEOUT 10000
+
+/**
+ * Timeout in msec for acquiring a lock of a session (used with a condition, so higher numbers could be required
+ * only in case of extreme concurrency).
+ */
+#define NC_SESSION_LOCK_TIMEOUT 500
+
+/**
+ * Timeout in msec for acquiring a lock of a session that is supposed to be freed.
+ */
+#define NC_SESSION_FREE_LOCK_TIMEOUT 1000
+
+/**
+ * Timeout in msec for a thread to wait for its turn to work with a pollsession structure.
+ */
+#define NC_PS_QUEUE_TIMEOUT 5000
+
+/**
+ * Time slept in msec if no endpoint was created for a running Call Home client.
+ */
+#define NC_CH_NO_ENDPT_WAIT 1000
+
+/**
+ * Time slept in msec between Call Home thread session idle timeout checks.
+ */
+#define NC_CH_THREAD_IDLE_TIMEOUT_SLEEP 1000
+
+/**
+ * Timeout in msec for a Call Home socket to establish its connection.
+ */
+#define NC_CH_CONNECT_TIMEOUT 500
+
+/**
+ * Number of sockets kept waiting to be accepted.
+ */
+#define NC_REVERSE_QUEUE 5
+
+/**
+ * Time slept in msec in each cycle of the client monitoring thread.
+ */
+#define NC_CLIENT_MONITORING_BACKOFF 200
+
+/**
+ * Timeout in msec for acquiring a lock of a client monitoring thread.
+ */
+#define NC_CLIENT_MONITORING_LOCK_TIMEOUT 500
+
+/**
+ * TLS key log file environment variable name.
+ */
+#define NC_TLS_KEYLOGFILE_ENV "SSLKEYLOGFILE"
+
+#define NC_VERSION_10_ENDTAG "]]>]]>"
+#define NC_VERSION_10_ENDTAG_LEN 6
+
 /**
  * Enumeration of diff operation types.
  */
@@ -65,12 +141,11 @@ enum nc_store_type {
 #include <curl/curl.h>
 #include <libssh/libssh.h>
 
-/* seconds */
+/**
+ * Timeout set for libssh (s).
+ */
 #define NC_SSH_TIMEOUT 10
 
-/* number of all supported authentication methods */
-#define NC_SSH_AUTH_COUNT 3
-
 /**
  * Enumeration of SSH public key formats.
  */
@@ -90,6 +165,14 @@ enum nc_privkey_format {
     NC_PRIVKEY_FORMAT_UNKNOWN   /**< Unknown format */
 };
 
+/**
+ * @brief Enumeration of SSH keyboard-interactive authentication methods.
+ */
+enum nc_kbdint_auth_method {
+    NC_KBDINT_AUTH_METHOD_NONE = 0,     /**< Keyboard-interactive authentication disabled. */
+    NC_KBDINT_AUTH_METHOD_SYSTEM        /**< Keyboard-interactive authentication is left to the system (PAM, shadow, etc.). */
+};
+
 /**
  * @brief A basic certificate.
  */
@@ -175,20 +258,21 @@ struct nc_auth_state {
  * @brief A server's authorized client.
  */
 struct nc_auth_client {
-    char *username;                         /**< Arbitrary username. */
+    char *username;                             /**< Arbitrary username. */
 
-    enum nc_store_type store;                    /**< Specifies how/where the client's public key is stored. */
+    enum nc_store_type store;                   /**< Specifies how/where the client's public key is stored. */
     union {
         struct {
-            struct nc_public_key *pubkeys;  /**< The client's public keys. */
-            uint16_t pubkey_count;          /**< The number of client's public keys. */
+            struct nc_public_key *pubkeys;      /**< The client's public keys. */
+            uint16_t pubkey_count;              /**< The number of client's public keys. */
         };
-        char *ts_ref;                       /**< Name of the referenced truststore key. */
+        char *ts_ref;                           /**< Name of the referenced truststore key. */
     };
 
-    char *password;                         /**< Client's password */
-    int kb_int_enabled;                     /**< Indicates that the client supports keyboard-interactive authentication. */
-    int none_enabled;                       /**< Implies that the client supports the none authentication method. */
+    char *password;                             /**< Client's password */
+    enum nc_kbdint_auth_method kbdint_method;   /**< Keyboard-interactive authentication method,
+                                                  *  may be extended by e.g. a union for each method when needed. */
+    int none_enabled;                           /**< Implies that the client supports the none authentication method. */
 };
 
 /**
@@ -327,6 +411,9 @@ struct nc_bind {
 
 #ifdef NC_ENABLED_SSH_TLS
 
+/* number of all supported authentication methods */
+#define NC_SSH_AUTH_COUNT 3
+
 struct nc_client_ssh_opts {
     char *knownhosts_path;  /**< path to known_hosts file */
     NC_SSH_KNOWNHOSTS_MODE knownhosts_mode; /**< implies whether to check known_hosts or not */
@@ -504,26 +591,25 @@ struct nc_ch_client_thread_arg {
 
 struct nc_server_opts {
     /* ACCESS locked - hello lock - separate lock to not always hold config_lock */
-    char **ignored_modules;         /**< Names of YANG modules that are not reported in the server <hello> message. */
-    uint16_t ignored_mod_count;
-    NC_WD_MODE wd_basic_mode;       /**< With-defaults basic mode of the server. */
-    int wd_also_supported;          /**< Bitmap of with-defaults modes that are also supported by the server. */
-    char **capabilities;
-    uint32_t capabilities_count;
-
-    char *(*content_id_clb)(void *user_data);   /**< Callback for generating content_id for ietf-yang-library data. */
-    void *content_id_data;
-    void (*content_id_data_free)(void *data);
-
     pthread_rwlock_t hello_lock;    /**< Needs to be held while the server <hello> message is being generated. */
 
-    /* ACCESS unlocked */
-    uint16_t idle_timeout;
+    char **ignored_modules;         /**< Names of YANG modules that are not reported in the server <hello> message. */
+    uint16_t ignored_mod_count;     /**< Number of ignored modules. */
+    NC_WD_MODE wd_basic_mode;       /**< With-defaults basic mode of the server. */
+    int wd_also_supported;          /**< Bitmap of with-defaults modes that are also supported by the server. */
+    char **capabilities;            /**< Array of server's capabilities. */
+    uint32_t capabilities_count;    /**< Number of server's capabilities. */
+
+    char *(*content_id_clb)(void *user_data);   /**< Callback for generating content_id for ietf-yang-library data. */
+    void *content_id_data;                      /**< Data passed to the content_id_clb callback. */
+    void (*content_id_data_free)(void *data);   /**< Callback to free the content_id_data. */
 
     /* ACCESS locked - options modified by YANG data/API - WRITE lock
      *               - options read when accepting sessions - READ lock */
     pthread_rwlock_t config_lock;
 
+    uint16_t idle_timeout;              /**< Idle timeout of the server sessions. */
+
 #ifdef NC_ENABLED_SSH_TLS
     char *authkey_path_fmt;             /**< Path to users' public keys that may contain tokens with special meaning. */
     char *pam_config_name;              /**< PAM configuration file name. */
@@ -532,182 +618,114 @@ struct nc_server_opts {
     void (*interactive_auth_data_free)(void *data);
 
     int (*user_verify_clb)(const struct nc_session *session);
+
+    struct nc_keystore keystore;        /**< Stored asymmetric and symmetric keys used by the server. */
+    struct nc_truststore truststore;    /**< Stored certificates and public keys used for authentication. */
 #endif /* NC_ENABLED_SSH_TLS */
 
-#ifdef NC_ENABLED_SSH_TLS
-    struct nc_keystore keystore;        /**< Server's keys/certificates. */
-    struct nc_truststore truststore;    /**< Server client's keys/certificates. */
-#endif /* NC_ENABLED_SSH_TLS */
-
-    /* ACCESS locked */
-    struct nc_bind *binds;
-    pthread_mutex_t bind_lock;          /**< To avoid concurrent calls of poll and accept on the bound sockets **/
     struct nc_endpt {
-        char *name;
+        char *name;                     /**< Identifier of the endpoint. */
 #ifdef NC_ENABLED_SSH_TLS
-        char *referenced_endpt_name;
+        char *referenced_endpt_name;    /**< Reference to another endpoint (used for client authentication). */
 #endif /* NC_ENABLED_SSH_TLS */
-        NC_TRANSPORT_IMPL ti;
-        struct nc_keepalives ka;
 
+        struct nc_keepalives ka;        /**< Keepalives configuration data. */
+
+        NC_TRANSPORT_IMPL ti;           /**< Transport implementation of the endpoint. */
         union {
 #ifdef NC_ENABLED_SSH_TLS
-            struct nc_server_ssh_opts *ssh;
-
-            struct nc_server_tls_opts *tls;
+            struct nc_server_ssh_opts *ssh;         /**< SSH transport options. */
+            struct nc_server_tls_opts *tls;         /**< TLS transport options. */
 #endif /* NC_ENABLED_SSH_TLS */
-            struct nc_server_unix_opts *unixsock;
+            struct nc_server_unix_opts *unixsock;   /**< UNIX socket transport options. */
         } opts;
-    } *endpts;
-    uint16_t endpt_count;
+    } *endpts;                          /**< Array of server's endpoints. */
+    uint16_t endpt_count;               /**< Number of server's endpoints. */
 
-    /* ACCESS locked, add/remove CH clients - WRITE lock ch_client_lock
-     *                modify CH clients - READ lock ch_client_lock + ch_client_lock */
     struct nc_ch_client {
-        char *name;
-        pthread_t tid;                                  /**< Call Home client's thread ID */
+        char *name;                                     /**< Identifier of the Call Home client. */
         struct nc_ch_client_thread_arg *thread_data;    /**< Data of the Call Home client's thread */
 
         struct nc_ch_endpt {
-            char *name;
+            char *name;                     /**< Identifier of the Call Home endpoint. */
 #ifdef NC_ENABLED_SSH_TLS
-            char *referenced_endpt_name;
+            char *referenced_endpt_name;    /**< Reference to another endpoint (used for client authentication). */
 #endif /* NC_ENABLED_SSH_TLS */
-            NC_TRANSPORT_IMPL ti;
 
-            char *src_addr;                             /**< IP address to bind to when connecting to a Call Home client. */
-            uint16_t src_port;                          /**< Port to bind to when connecting to a Call Home client. */
-            char *dst_addr;                             /**< IP address of the Call Home client. */
-            uint16_t dst_port;                          /**< Port of the Call Home client. */
+            char *src_addr;                 /**< IP address to bind to when connecting to a Call Home client. */
+            uint16_t src_port;              /**< Port to bind to when connecting to a Call Home client. */
+            char *dst_addr;                 /**< IP address of the Call Home client. */
+            uint16_t dst_port;              /**< Port of the Call Home client. */
 
-            int sock_pending;
-            struct nc_keepalives ka;
+            int sock_pending;               /**< Socket file descriptor of the pending connection to the Call Home client. */
+            struct nc_keepalives ka;        /**< Keepalives configuration data for the Call Home endpoint. */
 
+            NC_TRANSPORT_IMPL ti;           /**< Transport implementation of the Call Home endpoint. */
             union {
 #ifdef NC_ENABLED_SSH_TLS
-                struct nc_server_ssh_opts *ssh;
-
-                struct nc_server_tls_opts *tls;
+                struct nc_server_ssh_opts *ssh;     /**< SSH transport options for the Call Home endpoint. */
+                struct nc_server_tls_opts *tls;     /**< TLS transport options for the Call Home endpoint. */
 #endif /* NC_ENABLED_SSH_TLS */
             } opts;
-        } *ch_endpts;
-        uint16_t ch_endpt_count;
+        } *ch_endpts;                   /**< Array of Call Home endpoints. */
+        uint16_t ch_endpt_count;        /**< Number of Call Home endpoints. */
 
-        NC_CH_CONN_TYPE conn_type;
+        NC_CH_CONN_TYPE conn_type;      /**< Type of the Call Home connection. */
         struct {
-            uint16_t period;
-            time_t anchor_time;
-            uint16_t idle_timeout;
+            uint16_t period;            /**< Period of a periodic Call Home connection in seconds. */
+            time_t anchor_time;         /**< Anchor time of a periodic Call Home connection. */
+            uint16_t idle_timeout;      /**< Idle timeout of a periodic Call Home connection in seconds. */
         };
 
-        NC_CH_START_WITH start_with;
-        uint8_t max_attempts;
-        uint16_t max_wait;
+        NC_CH_START_WITH start_with;    /**< How to select the Call Home endpoint to connect to. */
+        uint8_t max_attempts;           /**< Maximum number of attempts to connect to the given Call Home endpoint. */
+        uint16_t max_wait;              /**< Maximum time to wait for a Call Home connection in seconds. */
         uint32_t id;
-        pthread_mutex_t lock;
-    } *ch_clients;
-    uint16_t ch_client_count;
-    pthread_rwlock_t ch_client_lock;
+    } *ch_clients;                      /**< Array of Call Home clients. */
+    uint16_t ch_client_count;           /**< Number of Call Home clients. */
 
 #ifdef NC_ENABLED_SSH_TLS
+    /**
+     * @brief Data for dispatching Call Home clients.
+     */
     struct nc_ch_dispatch_data {
-        nc_server_ch_session_acquire_ctx_cb acquire_ctx_cb;
-        nc_server_ch_session_release_ctx_cb release_ctx_cb;
-        void *ctx_cb_data;
-        nc_server_ch_new_session_cb new_session_cb;
-        void *new_session_cb_data;
+        nc_server_ch_session_acquire_ctx_cb acquire_ctx_cb;     /**< Acquiring libyang context callback. */
+        nc_server_ch_session_release_ctx_cb release_ctx_cb;     /**< Releasing libyang context callback. */
+        void *ctx_cb_data;                                      /**< Data passed to the callbacks above. */
+        nc_server_ch_new_session_cb new_session_cb;             /**< New session callback. */
+        void *new_session_cb_data;                              /**< Data passed to the new_session_cb callback. */
     } ch_dispatch_data;
-#endif /* NC_ENABLED_SSH_TLS */
 
-    /* ACCESS unlocked */
-    ATOMIC_T new_session_id;
-    ATOMIC_T new_client_id;
-
-#ifdef NC_ENABLED_SSH_TLS
-    /* ACCESS locked */
     struct {
-        pthread_t tid;                      /**< Thread ID of the certificate expiration notification thread. */
-        int thread_running;                 /**< Flag representing the runningness of the cert exp notification thread. */
-        pthread_mutex_t lock;               /**< Certificate expiration notification thread's data and cond lock. */
-        pthread_cond_t cond;                /**< Condition for the certificate expiration notification thread. */
+        pthread_t tid;              /**< Thread ID of the certificate expiration notification thread. */
+        int thread_running;         /**< Flag representing the runningness of the cert exp notification thread. */
+        pthread_mutex_t lock;       /**< Certificate expiration notification thread's data and cond lock. */
+        pthread_cond_t cond;        /**< Condition for the certificate expiration notification thread. */
 
         /**
          * @brief Intervals for certificate expiration notifications.
          */
         struct nc_interval {
-            struct nc_cert_exp_time anchor; /**< Lower bound of the given interval. */
-            struct nc_cert_exp_time period; /**< Period of the given interval. */
+            struct nc_cert_exp_time anchor;     /**< Lower bound of the given interval. */
+            struct nc_cert_exp_time period;     /**< Period of the given interval. */
         } *intervals;
-        int interval_count;                 /**< Number of intervals. */
+        int interval_count;                     /**< Number of intervals. */
     } cert_exp_notif;
+#endif /* NC_ENABLED_SSH_TLS */
+
+    /* ACCESS locked - bind_lock */
+    struct nc_bind *binds;              /**< Array of server's socket binds. The count is the same as endpt_count. */
+    pthread_mutex_t bind_lock;          /**< To avoid concurrent calls of poll and accept on the bound sockets. **/
 
     /* ACCESS unlocked */
+    ATOMIC_T new_session_id;
+
+#ifdef NC_ENABLED_SSH_TLS
+    /* ACCESS unlocked - set from env */
     FILE *tls_keylog_file;                  /**< File to log TLS secrets to. */
 #endif
 };
 
-/**
- * Sleep time in usec to wait between nc_recv_notif() calls.
- */
-#define NC_CLIENT_NOTIF_THREAD_SLEEP 10000
-
-/**
- * Timeout in msec for transport-related data to arrive (ssh_handle_key_exchange(), SSL_accept(), SSL_connect()).
- * It can be quite a lot on slow machines (waiting for TLS cert-to-name resolution, ...).
- */
-#define NC_TRANSPORT_TIMEOUT 10000
-
-/**
- * Timeout in msec for acquiring a lock of a session (used with a condition, so higher numbers could be required
- * only in case of extreme concurrency).
- */
-#define NC_SESSION_LOCK_TIMEOUT 500
-
-/**
- * Timeout in msec for acquiring a lock of a session that is supposed to be freed.
- */
-#define NC_SESSION_FREE_LOCK_TIMEOUT 1000
-
-/**
- * Timeout in msec for a thread to wait for its turn to work with a pollsession structure.
- */
-#define NC_PS_QUEUE_TIMEOUT 5000
-
-/**
- * Time slept in msec if no endpoint was created for a running Call Home client.
- */
-#define NC_CH_NO_ENDPT_WAIT 1000
-
-/**
- * Time slept in msec between Call Home thread session idle timeout checks.
- */
-#define NC_CH_THREAD_IDLE_TIMEOUT_SLEEP 1000
-
-/**
- * Timeout in msec for a Call Home socket to establish its connection.
- */
-#define NC_CH_CONNECT_TIMEOUT 500
-
-/**
- * Number of sockets kept waiting to be accepted.
- */
-#define NC_REVERSE_QUEUE 5
-
-/**
- * Time slept in msec in each cycle of the client monitoring thread.
- */
-#define NC_CLIENT_MONITORING_BACKOFF 200
-
-/**
- * Timeout in msec for acquiring a lock of a client monitoring thread.
- */
-#define NC_CLIENT_MONITORING_LOCK_TIMEOUT 500
-
-/**
- * TLS key log file environment variable name.
- */
-#define NC_TLS_KEYLOGFILE_ENV "SSLKEYLOGFILE"
-
 /**
  * @brief Type of the session
  */
@@ -724,9 +742,6 @@ enum nc_version {
     NC_VERSION_11 = 1   /**< NETCONF 1.1 - RFC 6241, 6242 */
 };
 
-#define NC_VERSION_10_ENDTAG "]]>]]>"
-#define NC_VERSION_10_ENDTAG_LEN 6
-
 /**
  * @brief Container to serialize RPC messages
  */
@@ -1188,6 +1203,22 @@ NC_MSG_TYPE nc_connect_callhome(const char *host, uint16_t port, NC_TRANSPORT_IM
 
 #ifdef NC_ENABLED_SSH_TLS
 
+/**
+ * @brief Dispatch a thread connecting to a listening NETCONF client and creating Call Home sessions.
+ *
+ * @param[in] client_name Existing client name.
+ * @param[in] acquire_ctx_cb Callback for acquiring new session context.
+ * @param[in] release_ctx_cb Callback for releasing session context.
+ * @param[in] ctx_cb_data Arbitrary user data passed to @p acquire_ctx_cb and @p release_ctx_cb.
+ * @param[in] new_session_cb Callback called for every established session on the client.
+ * @param[in] new_session_cb_data Arbitrary user data passed to @p new_session_cb.
+ * @param[in] config_locked Whether the server configuration is locked.
+ * @return 0 if the thread was successfully created, -1 on error.
+ */
+int _nc_connect_ch_client_dispatch(const char *client_name, nc_server_ch_session_acquire_ctx_cb acquire_ctx_cb,
+        nc_server_ch_session_release_ctx_cb release_ctx_cb, void *ctx_cb_data, nc_server_ch_new_session_cb new_session_cb,
+        void *new_session_cb_data, int config_locked);
+
 /**
  * @brief Accept a server Call Home connection on a socket.
  *
