@@ -219,8 +219,9 @@ void nc_server_tls_set_verify_wrap(void *tls_cfg, struct nc_tls_verify_cb_data *
  * @brief Set TLS client's verify flags.
  *
  * @param[in] tls_cfg TLS configuration.
+ * @return 0 on success, 1 on error.
  */
-void nc_client_tls_set_verify_wrap(void *tls_cfg);
+int nc_client_tls_set_verify_wrap(void *tls_cfg);
 
 /**
  * @brief Verify the certificate.
@@ -451,11 +452,10 @@ int nc_tls_init_ctx_wrap(void *cert, void *pkey, void *cert_store, void *crl_sto
  * @brief Setup a TLS configuration from a TLS context.
  *
  * @param[in] tls_ctx TLS context.
- * @param[in] side Side of the TLS connection.
  * @param[in,out] tls_cfg TLS configuration.
  * @return 0 on success, non-zero on fail.
  */
-int nc_tls_setup_config_from_ctx_wrap(struct nc_tls_ctx *tls_ctx, int side, void *tls_cfg);
+int nc_tls_setup_config_from_ctx_wrap(struct nc_tls_ctx *tls_ctx, void *tls_cfg);
 
 /**
  * @brief Get the error code from a TLS session's verification.
